@@ -3,16 +3,20 @@ import base64
 import json
 import os
 import subprocess
+import uuid
 
 def main():
-    server_ip = os.getenv('SERVER_IP')
+    server_ip = os.getenv('MAXPLOTLIB_SERVER_IP')
+
+    new_uuid = uuid.uuid4()
 
     parser = argparse.ArgumentParser()
     parser.add_argument('prompt', type=str, help='Prompt describing the plot to generate.')
-    parser.add_argument('--output', type=str, default='maxplotlib_output', help='Output directory to save the image.')
+    parser.add_argument('--output', type=str, default=f'maxplotlib_output_{new_uuid}', help='Output directory to save the image.')
     args = parser.parse_args()
 
     os.makedirs(args.output, exist_ok=True)
+    os.makedirs(f'{args.output}', exist_ok=True)
 
     curl_command = [
         'curl', '-s', '-X', 'POST',
